"""
Generates a training set for training an abnormality detection model from the
data directory specified by the first argument.  This uses the representations
obtained from a device classifier model specified by the second argument to
condition on
"""

import sys
import os
import json
import logging
import pickle

import numpy as np
from .RandomForestModel import RandomForestModel
from .pcap_utils import clean_session_dict
from .pcap_utils import get_source
from .pcap_utils import featurize_session

logging.basicConfig(level=logging.INFO)

# Get model info from config
with open('opts/config.json') as config_file:
    config = json.load(config_file)
    state_size = config['state size']
    duration = config['duration']
    time_const = config['time constant']

def average_representation(rep, timestamp, prev_rep, prev_time):
    """
    Computes the new moving average representation from a single input
    """

    # If no previous info, the average is just the input
    if prev_rep is None or prev_time is None:
        return rep, timestamp

    #Otherwise, compute the moving average
    delta_t = timestamp.timestamp() - prev_time.timestamp()
    alpha = 1 - np.exp(-delta_t/time_const)
    new_rep = prev_rep + alpha*(rep - prev_rep)

    return new_rep, timestamp

def create_dataset(
                    data_dir,
                    model_path='/models/OneLayerModel.pkl',
                    label=None
                  ):
    logger = logging.getLogger(__name__)

    # Load the model
    logger.debug("Loading model")
    model = RandomForestModel(duration=None, hidden_size=None)
    model.load(model_path)

    # Get all the pcaps in the training directory
    logger.debug("Getting pcaps")
    pcaps = []
    try:
        name, ext = os.path.splitext(data_dir)
        if ext == '.pcap':
            pcaps.append(data_dir)
    except:
        pass

    for dirpath, dirnames, filenames in os.walk(data_dir):
        for filename in filenames:
            name, ext = os.path.splitext(filename)
            if ext == '.pcap':
                pcaps.append(os.path.join(dirpath,filename))

    # Get and store the representations using the supplied model
    # Representations will be computed separately for each pcap
    representations = {}
    for pcap in pcaps:
        logger.debug("Working on %s", pcap)
        reps, _, timestamps, _, _ = model.get_representation(
                                                            pcap,
                                                            mean=False
                                                                    )
        sessions = model.sessions
        source_address = get_source(sessions)

        # Compute the mean representations
        prev_rep = None
        prev_time = None
        model_outputs = {}

        if timestamps is not None:
            for i, timestamp in enumerate(timestamps):
                rep = reps[i]
                new_rep, time = average_representation(
                                                       rep,
                                                       timestamp,
                                                       prev_rep,
                                                       prev_time
                                                      )
                preds = model.classify_representation(new_rep)
                if label is not None:
                    preds = [(p[0],0) for p in preds if p[0] != label]
                    preds.append((label,1))

                model_outputs[timestamp] = {
                                            "classification": list(preds),
                                            "representation": list(rep),
                                            "mean representation": list(new_rep)
                                           }
                prev_rep, prev_time = new_rep, time

        # Clean the sessions and merge them into a single session dict
        session_rep_pairs = []
        source = get_source(sessions, address_type='IP')
        for session_dict in sessions:
            for key, value in session_dict.items():
                session_info = featurize_session(key, value, source=source)

                first_time = value[0][0].timestamp()
                prior_time = None
                for timestamp in timestamps:
                    time = timestamp.timestamp()
                    if first_time > time:
                        prior_time = timestamp
                if prior_time == None:
                    prior_time = timestamps[0]

                pair = {
                        "model outputs": model_outputs[prior_time],
                        "session info": session_info,
                        "key": key
                       }
                if session_info is not None:
                    session_rep_pairs.append(pair)

        representations[pcap] = session_rep_pairs
    byte_size = sys.getsizeof(pickle.dumps(representations))
    logger.debug(
                "created training data of size %f mb",
                round(byte_size/1000000, 3)
               )

    return representations



"""
Contains iterator class for generating training batches from a canned dataset
"""

import pickle
import numpy as np
from sklearn.model_selection import train_test_split

class BatchIterator:
    def __init__(
                 self,
                 data_input,
                 labels,
                 batch_size=64,
                 seq_len=116,
                 num_chars=16,
                 perturb_types="all",
                 seed=0
                ):
        """
        Initialize the iterator with specified hyperparameters and load the
        data from the specified path
        """

        self.rng = np.random.RandomState(seed)

        self.data_input = data_input
        self.labels = sorted(labels)
        self.classification_length = len(labels)

        self.data = None
        self.train_sessions_by_length = None
        self.train_labels_by_length = None
        self.vala_sessions_by_length = None
        self.vala_labels_by_length = None
        self.test_sessions_by_length = None
        self.test_labels_by_length = None
        self.load_data()

        self.seq_len = seq_len
        self.num_chars = num_chars
        self.perturb_types = perturb_types

    def load_data(self):
        """
        Handles loading the data into the correct format
        """
        if type(self.data_input) is dict:
            print("Loading data from dict")
            self.data = self.data_input
        else:
            print("Loading data from disk")
            with open(self.data_input,'rb') as handle:
                data = pickle.load(handle)
            self.data = data

        train_sessions_by_length = [{} for i in range(9)]
        train_labels_by_length = [{} for i in range(9)]
        vala_sessions_by_length = [{} for i in range(9)]
        vala_labels_by_length = [{} for i in range(9)]
        test_sessions_by_length = [{} for i in range(9)]
        test_labels_by_length = [{} for i in range(9)]

        for key, value in self.data.items():
            for session in value:
                session_length = len(session["packets"])
                model_output = session['model outputs']
                class_array = [c for c,p in model_output['classification']]
                decision = class_array[0]

                # Randomly partition the data
                split = np.random.choice([1, 2, 3], p=[0.8, 0.1, 0.1])
                if split == 1:
                    session_dict = train_sessions_by_length[session_length]
                    label_dict = train_labels_by_length[session_length]
                if split == 2:
                    session_dict = vala_sessions_by_length[session_length]
                    label_dict = vala_labels_by_length[session_length]
                if split == 3:
                    session_dict = test_sessions_by_length[session_length]
                    label_dict = test_labels_by_length[session_length]

                if decision not in label_dict:
                    label_dict[decision] =0
                if decision not in session_dict:
                    session_dict[decision] = []

                session_dict[decision].append(session)
                label_dict[decision] += 1

        self.train_sessions_by_length = train_sessions_by_length
        self.train_labels_by_length = train_labels_by_length
        self.vala_sessions_by_length = vala_sessions_by_length
        self.vala_labels_by_length = vala_labels_by_length
        self.test_sessions_by_length = test_sessions_by_length
        self.test_labels_by_length = test_labels_by_length

    def gen_data(self, length=8, split='train', batch_size=64, perturb=False):
        """
        Generates perturbed or unperturbed batches
        """
        X = np.zeros((batch_size,length,self.seq_len,self.num_chars))
        L = np.zeros((batch_size,self.classification_length))
        Y = np.zeros((batch_size,1))
        chosen_mods = []
        if perturb is True:
            Y = np.ones((batch_size,1))

        hex_str = '0123456789abcdef'
        mod_types = ['label_swap', 'packet_swap', 'duplicate']
        if self.perturb_types != "all":
            mod_types = self.perturb_types
        if length <= 1:
            mod_types = ['label_swap']

        if split == 'train':
            sessions_by_length = self.train_sessions_by_length
        if split == 'vala':
            sessions_by_length = self.vala_sessions_by_length
        if split == 'test':
            sessions_by_length = self.test_sessions_by_length

        all_keys = list(sessions_by_length[length].keys())
        chosen_keys = np.random.choice(
                                        all_keys,
                                        size=batch_size,
                                        replace=True
                                      )

        for i, key in enumerate(chosen_keys):
            if perturb is True:
                mod_type = np.random.choice(mod_types)
            else:
                mod_type = None
            chosen_mods.append(mod_type)
            sessions = sessions_by_length[length]
            session_id = np.random.choice(len(sessions[key]))
            session = sessions[key][session_id]
            model_outputs = session["model outputs"]

            classification = model_outputs['classification']
            classification = sorted(classification, key=lambda x: x[0])

            # Compute the label vectors using the supplied labels
            L_class = np.zeros(self.classification_length)
            indicies = [self.labels.index(c) for c, p in classification]
            probas = [p for c,p in classification]
            if mod_type == 'label_swap':
                # Shuffle the indicies for a label swap
                np.random.shuffle(indicies)
            for i, idx in enumerate(indicies):
                L_class[idx] = probas[i]

            L[i] = L_class

            packets = session["packets"]
            for j, packet in enumerate(packets):
                for k, char in enumerate(packet[1]):
                    if k < self.seq_len:
                        X[i,j,k,hex_str.index(char)] = 1
            if mod_type == 'packet_swap':
                id_1, id_2 = np.random.choice(range(length), size=2)
                X[i,id_1,:,:], X[i,id_2,:,:] = X[i,id_2,:,:], X[i,id_1,:,:]
            if mod_type == 'duplicate':
                id_1 = np.random.choice(range(length-1))
                id_2 = id_1 + 1
                X[i,id_2,:,:] = X[i,id_1,:,:]


        return X, L, Y, chosen_mods

    def gen_batch(self, split='train', length=8, batch_size=64):
        """
        Generate mixed batches of perturbed and unperturbed data
        """
        X_n, L_n, Y_n, c_n = self.gen_data(
                                           length=8,
                                           split=split,
                                           batch_size=batch_size//2
                                          )
        X_a, L_a, Y_a, c_a = self.gen_data(
                                           length=8,
                                           split=split,
                                           batch_size=batch_size//2,
                                           perturb=True
                                          )
        X = np.concatenate((X_n,X_a), axis=0)
        L = np.concatenate((L_n,L_a), axis=0)
        Y = np.concatenate((Y_n,Y_a), axis=0)
        c = c_n + c_a
        return X, L, Y, c



import os
import sys
import json
import logging
import pickle
import numpy as np
import tensorflow as tf
from .SoSmodel import SoSModel
from .session_sequence import create_dataset
from .session_iterator import BatchIterator
import time

logging.basicConfig(level=logging.INFO)
tf.logging.set_verbosity(tf.logging.ERROR)

# Load info from config
with open('opts/config.json') as config_file:
    config = json.load(config_file)
    rnn_size = config['rnn size']
    labels = config['labels']

if __name__ == '__main__':
    logger = logging.getLogger(__name__)
    # Path to training data
    data_dir = sys.argv[1]
    # Create the training data
    if len(sys.argv) == 3:
        data = create_dataset(data_dir)
        write_dir = sys.argv[2]
        logger.info("Saving data to %s", write_dir)
        with open(write_dir, 'wb') as handle:
            pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        data = data_dir

    logger.info("Loaded training data")
    # Create an iterator
    iterator = BatchIterator(
                             data,
                             labels,
                             perturb_types=['random data','port swap', 'direction_swap']
                            )
    logger.info("Created iterator")
    rnnmodel = SoSModel(rnn_size=100)
    logger.info("Created model")
    try:
        rnnmodel.load('/models/SoSmodel')
        logger.info("Loaded model")
    except:
        rnnmodel.initialize()
        logger.info("Initialized model")

    X_v, L_v, Y_v = iterator.gen_batch(
                                            split='validation',
                                            batch_size=64
                                         )

    cost = rnnmodel.get_cost(X_v,L_v,Y_v)
    out = rnnmodel.get_output(X_v, L_v)

    logger.info("Initial validation cost: %s",np.mean(cost))
    min_cost = cost
    last_save = 0
    for i in range(100000):
        tick = time.clock()
        X,L,Y = iterator.gen_batch(
                                    split='train',
                                    batch_size=64
                                  )
        tock = time.clock()
        _ = rnnmodel.train_on_batch(X,L,Y)
        if (i+1)%100 == 0:
            cost = rnnmodel.get_cost(X_v,L_v,Y_v)
            logger.info("Validation cost after  %s batches: %s",i,cost)
            if cost < min_cost:
                min_cost = cost
                rnnmodel.save('/models/SoSmodel')
                last_save = 0
                logger.info('Saving model at validation cost %s', cost)
            else:
                last_save += 100
        if last_save > 1000:
            logger.info("No improvement after 1000 iterations. Stopping.")
            break



import numpy as np
import pickle as pickle
from .reader import sessionizer
from .featurizer import extract_features

from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPClassifier
from sklearn.metrics import f1_score

from .training_utils import read_data
from .training_utils import select_features


class OneLayerModel:
    def __init__(self, duration, hidden_size=None, labels=None):
        '''
        Initializes a model with a single hidden layer.  Features are
        aggregated over the time specified by the duration and the hidden
        layer size is a hyperparameter set at initialization.

        Args:
            duration: Time duration to aggregate features for
        '''

        self.duration = duration
        self.hidden_size = hidden_size

        self.means = None
        self.stds = None
        self.feature_list = None
        self.model = None
        self.labels = labels

        self.sessions = None

    def _augment_data(self, X, y):
        '''
        Augments the data with randomly permuted samples. Added samples are
        labeled Unknown.

        Args:
            X: Numpy 2D array containing data to augment
            y: Numpy 1D array of labels to augment

        Returns:
            X_aug: Numpy 2D array containing augmented data
            y_aug: Numpy 1D array containing labels of augmented data
        '''

        # Randomly permute the inputs, and label the permutations Unknown
        X_permed = np.copy(X)
        for i in range(X_permed.shape[1]):
            np.random.shuffle(X_permed[:,i])

        y_permed = [self.labels.index("Unknown")]*X_permed.shape[0]
        y_permed = np.stack(y_permed)

        X_aug = np.concatenate((X, X_permed), axis=0)
        y_aug = np.concatenate((y, y_permed), axis=0)

        return X_aug, y_aug

    def get_features(self, filepath, source_ip=None):
        '''
        Reads a pcap specified by the file path and returns an array of the
        computed model inputs

        Args:
            filepath: Path to pcap to compute features for

        Returns:
            features: Numpy 2D array containing features for each time bin
            timestamp: datetime of the last observed packet
        '''

        # Read the capture into a feature array
        X = []
        timestamps = []
        binned_sessions = sessionizer(filepath, duration=self.duration)
        self.sessions = binned_sessions

        if len(binned_sessions) is 0:
            return None, None, None, None

        for session_dict in binned_sessions:
            if len(session_dict) > 0:
                if source_ip is None:
                    feature_list, source_ip, other_ips = extract_features(
                                                                    session_dict
                                                                         )
                else:
                    feature_list, _, other_ips = extract_features(
                                                      session_dict,
                                                      capture_source=source_ip
                                                                 )
                X.append(feature_list)
                last_packet = list(session_dict.items())[-1]
                timestamps.append(last_packet[1][0][0])

        if len(X) == 0:
            return None, None, None, None

        full_features = np.stack(X)

        # Mean normalize the features
        full_features -= np.expand_dims(self.means, 0)
        full_features /= np.expand_dims(self.stds, 0)
        features = full_features[:, self.feature_list]

        '''
        last_packet = list(binned_sessions[-1].items())[-1]
        timestamp = last_packet[1][0][0]
        '''

        return features, source_ip, timestamps, other_ips


    def train(self, data_dir):
        '''
        Trains a single layer model on the data contained in the specified
        directory.  Labels found in the directory are augmented with an
        unknown label.

        Args:
            data_dir: Directory containing the training data
        '''

        print("Reading data")
        # First read the data directory for the features and labels
        X_all, y_all, new_labels = read_data(
                                              data_dir,
                                              duration=self.duration,
                                              labels=self.labels
                                            )
        self.labels = new_labels

        print("Making data splits")
        # Split the data into training, validation, and testing sets
        X_train, X_test, y_train, y_test = train_test_split(
                                                            X_all,
                                                            y_all,
                                                            test_size=0.2,
                                                            random_state=0
                                                           )

        print("Normalizing features")
        # Mean normalize the features, saving the means and variances
        self.means = X_train.mean(axis=0)
        self.stds = X_train.std(axis=0)
        # Set the zero standard deviations to 1
        zero_stds = self.stds <= 1
        self.stds[zero_stds] = 1
        # Apply the mean normalization transformation to the training dataj
        X_normed = X_train - np.expand_dims(self.means, 0)
        X_normed /= np.expand_dims(self.stds, 0)

        print("Doing feature selection")
        # Select the relevant features from the training set
        self.feature_list = select_features(X_normed, y_train)
        print(self.feature_list)

        # If hidden size wasn't specified, default to the mean of the number
        # of features and the size of the label space
        if self.hidden_size is None:
            self.hidden_size = int(1/2*(
                                        len(self.labels) + \
                                        len(self.feature_list)
                                       )
                                  )

        # Augment the data with randomly permuted samples
        X_aug, y_aug = self._augment_data(X_normed, y_train)

        # Fit the one layer model to the augmented training data
        X_input = X_aug[:, self.feature_list]
        self.model = MLPClassifier(
                                    (self.hidden_size),
                                    alpha=0.1,
                                    activation='relu',
                                    max_iter=1000
                                  )

        self.model.fit(X_input, y_aug)

        # Evaulate the model on the augmented test data
        X_test_input = X_test - np.expand_dims(self.means, 0)
        X_test_input /= np.expand_dims(self.stds, 0)
        X_test_aug, y_test_aug = self._augment_data(X_test_input, y_test)
        predictions = self.model.predict(X_test_aug[:, self.feature_list])
        print("F1 score:",
                f1_score(y_test_aug, predictions, average='weighted'))

    def predict(self, filepath, source_ip=None):
        '''
        Read a capture file from the specified path and make a prediction
        of the source.

        Args:
            filepath: Path of capture file to read

        Returns:
            prediction: list of tuples formatted as (source, probability)
        '''

        features, _, _, _ = self.get_features(filepath, source_ip=source_ip)

        if features is None:
            return None
        predictions = self.model.predict_proba(features)
        mean_predictions = np.mean(predictions, axis=0)

        prediction = [
                      (self.labels[i], prob)
                      for i, prob in enumerate(mean_predictions)
                     ]
        prediction = sorted(prediction, key=lambda x: x[1], reverse=True)
        return prediction

    def get_representation(self, filepath, mean=True, source_ip=None):
        '''
        Computes the mean hidden representation of the input file.

        Args:
            filepath: Path of capture file to represent
            mean: If true(default), averages all the representations into one

        Returns:
            representation:  representation vector of the input file
        '''

        features, source_ip, timestamp, other_ips = self.get_features(
                                                           filepath,
                                                           source_ip=source_ip,
                                                                     )
        if features is None:
            return None, None, None, None, None

        L1_weights = self.model.coefs_[0]
        L1_biases = self.model.intercepts_[0]
        representation = np.maximum(
                                     np.matmul(features, L1_weights)+L1_biases,
                                     0
                                   )

        mean_rep = np.mean(representation, axis=0)

        L2_weights = self.model.coefs_[1]
        L2_biases = self.model.intercepts_[1]
        probabilities = np.matmul(representation, L2_weights) + L2_biases
        probabilities = np.exp(probabilities)
        probabilities /= np.expand_dims(np.sum(probabilities, axis=1), axis=1)
        probabilities = np.mean(probabilities, axis=0)
        prediction = [
                        (self.labels[i], prob)
                        for i, prob in enumerate(probabilities)
                     ]
        prediction = sorted(prediction, key=lambda x: x[1], reverse=True)

        if mean:
            representation = mean_rep
            timestamp = timestamp[-1]

        return representation, source_ip, timestamp, prediction, other_ips

    def classify_representation(self, representation):
        '''
        Takes in a representation and produces a classification
        '''
        L2_weights = self.model.coefs_[1]
        L2_biases = self.model.intercepts_[1]
        probabilities = np.matmul(representation, L2_weights) + L2_biases
        probabilities = np.exp(probabilities)
        probabilities /= np.sum(probabilities)
        prediction = [
                        (self.labels[i], prob)
                        for i, prob in enumerate(probabilities)
                     ]
        prediction = sorted(prediction, key=lambda x: x[1], reverse=True)

        return prediction

    def save(self, save_path):
        '''
        Saves the model to the specified file path

        Args:
            save_path: Path to store the saved model at.
        '''

        model_attributes = {
                            'duration': self.duration,
                            'hidden_size': self.hidden_size,
                            'means': self.means,
                            'stds': self.stds,
                            'feature_list': self.feature_list,
                            'model': self.model,
                            'labels': self.labels
                           }

        with open(save_path, 'wb') as handle:
            pickle.dump(model_attributes, handle)

    def load(self, load_path):
        '''
        Load the model parameters from the specified path.

        Args:
            load_path: Path to load the model parameters from
        '''

        with open(load_path, 'rb') as handle:
            model_attributes = pickle.load(handle)

        self.duration = model_attributes['duration']
        self.hidden_size = model_attributes['hidden_size']
        self.means = model_attributes['means']
        self.stds = model_attributes['stds']
        self.feature_list = model_attributes['feature_list']
        self.model = model_attributes['model']
        self.labels = model_attributes['labels']



import json
import os
import subprocess
from collections import OrderedDict
import datetime

def parse_packet_head(line):
    '''
    Parses the head of the packet to get the key tuple which contains
    the flow level data

    Args:
        line: Header line from tcpdump

    Returns:
        key: Tuple key which contains packet info
    '''

    # Split the header line into its components
    data = line.decode('utf8')
    data = data.split(' ')

    # Only generate a key if this packet contains IP information
    if len(data) < 2:
        return None

    # Parse out the date and time the packet was seen
    date_str = data[0] + ' ' + data[1]
    date = datetime.datetime.strptime(date_str, '%Y-%m-%d %H:%M:%S.%f')

    # Parse out the source and destination addresses and ports
    source_data = data[3].split('.')
    destination_data = data[5].split('.')
    destination_port = '0'
    source_port = '0'

    # ipv4 packet
    if data[2] == 'IP':
        # if TCP or UDP set port
        if len(source_data) >= 5:
            source_port = source_data[4]

        source_str = '.'.join(source_data[0:4]) + ':' + source_port
        if len(destination_data) < 5:
            destination_str = '.'.join(destination_data[0:4])[0:-1] \
                              + ':' \
                              + destination_port
        else:
            destination_port = destination_data[4][0:-1]
            destination_str = '.'.join(destination_data[0:4]) \
                              + ':' \
                              + destination_port
    # ipv6 packet
    elif data[2] == 'IP6':
        # if TCP or UDP set port
        if len(source_data) >= 2:
            source_port = source_data[1]

        source_str = source_data[0] + ':' + source_port
        if len(destination_data) < 2:
            destination_str = destination_data[0][0:-1] + ':' + destination_port
        else:
            destination_port = destination_data[1][0:-1]
            destination_str = destination_data[0] \
                              + ':' \
                              + destination_port
    else:
        return None

    return date, source_str, destination_str

def parse_packet_data(line):
    '''
    Parses the hex data from a line in the packet and returns it as a
    string of characters in 0123456789abcdef.

    Args:
        line: Hex output from tcpdump

    Returns:
        packet_data: String containing the packet data
    '''
    raw_data = line.decode('utf-8')
    try:
        _, data = raw_data.split(':', 1)
    except ValueError:
        return None
    packet_data = data.strip().replace(' ' ,'')

    return packet_data

def packetizer(path):
    '''
    Reads a pcap specified by the path and parses out the packets.
    Packets will be stored with a tuple key formatted as follows:
    (datetime, sIP:sPort, dIP:dPort, protocol, length)

    Args:
        path: Path to pcap to read

    Returns:
        packet_dict: Dictionary of packets with keys formatted as above
    '''

    # Read get the pcap info with tcpdump
    FNULL = open(os.devnull, 'w')
    proc = subprocess.Popen(
                            'tcpdump -nn -tttt -xx -r' + path,
                            shell=True,
                            stdout=subprocess.PIPE,
                            stderr=FNULL
                           )
    head = None
    packet_dict = OrderedDict()
    # Go through all the lines of the output
    for line in proc.stdout:
        if not line.startswith(b'\t'):
            head = parse_packet_head(line)
            if head is not None:
                packet_dict[head] = ''
        elif head is not None:
            data = parse_packet_data(line)
            if data is not None:
                packet_dict[head] += data
    return packet_dict

def sessionizer(path, duration=None, threshold_time=None):
    '''
    Reads a pcap specified by the path and parses out the sessions.
    Sessions are defined as flows with matching sourceIP:sourcePort
    and destinationIP:destinationPorts. The sessions can also be binned
    in time according to the optional duration parameter.

    Args:
        path: Path to pcap to read
        duration: Duration of session bins. None uses a single bin for
                  the entire pcap.

    Returns:
        session_dict: Dictionary of sessions with keys as tuples of
                      (sourceIP:sourcePort, destIP:destPort)
    '''

    # Get the packets from the pcap
    packet_dict = packetizer(path)

    # Go through the packets one by one and add them to the session dict
    sessions = []
    start_time = None
    working_dict = None

    first_packet_time = None
    session_starts = OrderedDict()

    # Get threshold time from config
    if threshold_time is None:
        try:
            with open('opts/config.json', 'r') as config_file:
                config = json.load(config_file)
                threshold_time  = config['session threshold']
        except Exception as e:
            threshold_time = 120

    for head, packet in packet_dict.items():
        time = head[0]

        # Get the time of the first observed packet
        if first_packet_time is None:
            first_packet_time = time

        # Start off the first bin when the first packet is seen
        if start_time is None:
            start_time = time
            working_dict = OrderedDict()

        # If duration has been specified, check if a new bin should start
        if duration is not None:
            if (time-start_time).total_seconds() >= duration:
                sessions.append(working_dict)
                working_dict = OrderedDict()
                start_time = time

        # Add the key to the session dict if it doesn't exist
        key_1 = (head[1], head[2])
        key_2 = (head[2], head[1])

        # Select the appropriate ordering
        if key_2 in working_dict:
            key = key_2
        if key_1 in working_dict:
            key = key_1

        if key_1 not in working_dict and key_2 not in working_dict:
            if key_1 not in session_starts and key_2 not in session_starts:
                session_starts[key_1] = time

            if key_1 in session_starts:
                session_start = session_starts[key_1]
            if key_2 in session_starts:
                session_start = session_starts[key_2]

            key = key_1
            if (session_start - first_packet_time).total_seconds() > threshold_time:
                working_dict[key] = []

        # Add the session to the session dict if it's start time is after
        # the cutoff
        if key in working_dict:
            working_dict[key].append((head[0],packet))

    if duration is not None and working_dict is not None:
        if len(working_dict) > 0:
            sessions.append(working_dict)
    if duration is None:
        sessions.append(working_dict)
    return sessions



'''
Contains utilities required for parsing pcaps into model training features
'''
import os
import json
import numpy as np

from sklearn.model_selection import cross_val_score
from sklearn.decomposition import PCA

from sklearn.linear_model import RandomizedLogisticRegression
from sklearn.linear_model import LogisticRegression

from .reader import sessionizer
from .pcap_utils import get_source
from .featurizer import extract_features

def read_data(data_dir, duration=None, labels=None):
    '''
    Reads all the data in the specified directory and parses it into
    a feature array and a label array.

    Args:
        data_dir: path to the directory that contains the training data
        duration: Time window to compute feature information
        labels: List containing labels to use

    Returns:
        X: numpy 2D array that contains the (high dimensional) features
        y: numpy 1D array that contains the labels for the features in X
        new_labels: Reordered labels used in training
    '''
    X = []
    y = []
    assigned_labels = []

    # Get all the files in the directory
    files = []
    with open('opts/label_assignments.json') as handle:
        label_assignments = json.load(handle)

    for dirpath, dirnames, filenames in os.walk(data_dir):
        for file in filenames:
            _, ext = os.path.splitext(file)
            if ext == '.pcap':
                files.append(os.path.join(dirpath,file))
    # Go through all the files in the directory
    for filename in files:
        # Extract the label from the filename
        name = os.path.split(filename)[1]
        name = name.split('-')[0]
        if name in label_assignments:
            label = label_assignments[name]
            if label not in labels: label = 'Unknown'
        else:
            label = 'Unknown'
        if label not in assigned_labels:
            assigned_labels.append(label)

        print("Reading", filename,"as",label)
        # Bin the sessions with the specified time window
        binned_sessions = sessionizer(
                                       filename,
                                       duration=duration
                                     )
        # Get the capture source from the binned sessions
        capture_source = get_source(binned_sessions)

        # For each of the session bins, compute the  full feature vectors
        for session_dict in binned_sessions:
            features, _, _ = extract_features(
                                            session_dict,
                                            capture_source=capture_source
                                             )

            # Store the feature vector and the labels
            X.append(features)
            y.append(assigned_labels.index(label))

        # Update the labels to reflect the new assignments
        new_labels = assigned_labels + \
                     [l for l in labels if l not in assigned_labels]

    return np.stack(X), np.stack(y), new_labels

def select_features(X, y):
    '''
    Select the relevant features from X that are useful for predicting
    the labels in y.

    Args:
        X: numpy 2D array containing input features
        y: numpy 1D array containing labels

    Returns:
        feature_list: List of indices of the selected important features
    '''

    # Get the selection model (stability selection)
    selection_model = RandomizedLogisticRegression(random_state=0)
    selection_model.fit(X, y)

    # Use a cross validated logistic regression to choose the importance
    # threshold at which a feature is included
    step_size = 50
    max_weight = int(max(selection_model.scores_)) + 1
    trial_thresholds = [i/step_size for i in range(1,max_weight*step_size + 1)]
    threshold = 0
    max_score = 0
    for trial in trial_thresholds:
        selected_features = [i
                        for i, score in enumerate(selection_model.scores_)
                        if score > trial]
        if len(selected_features) > 0:
            X_reduced = X[:, selected_features]
            model = LogisticRegression(
                                       multi_class='multinomial',
                                       class_weight='balanced',
                                       solver='newton-cg',
                                       random_state=0,
                                       max_iter=1000
                                      )
            scores = cross_val_score(model, X_reduced, y, cv=5)
            score = scores.mean()
            if score >= max_score:
                max_score = score
                threshold = trial/step_size

    importance = {i:s for i,s in enumerate(selection_model.scores_)}
    return [i for i, score in enumerate(selection_model.scores_)
            if score > threshold]

def whiten_features(X):
    '''
    Fits the witening transformation for the features X. and returns the
    associated matrix.

    Args:
        X: numpy 2D array containing features

    Returns:
        whitening_transformation: Transformation to whiten features
    '''

    # Use PCA to create a whitening transformation fit to the training set
    whitening_transformation = PCA(whiten=False)
    whitening_transformation.fit(X)

    return whitening_transformation

def choose_regularization(X, y):
    '''
    Chooses a value for the regularization parameter using grid search and
    cross validation.

    Args:
        X: numpy 2D array of model inputs
        y: numpy 1D array of labels

    Returns:
        C: Selected value of the regulatization coefficient
    '''

    # Set up the grid search
    max_C, step_size = 10, 5
    best_score, C = 0, 0
    trial_Cs = [i/step_size for i in range(1, max_C*step_size + 1)]

    # Grid search with cross validation to get C
    for trial in trial_Cs:
        model = LogisticRegression(
                                    C=trial,
                                    multi_class='multinomial',
                                    solver='newton-cg',
                                    class_weight='balanced',
                                    random_state=0,
                                    max_iter=1000
                                  )
        scores = cross_val_score(model, X, y, cv=10)
        score = scores.mean()
        if score > best_score:
            best_score = score,
            C = trial

    return C



import functools
import numpy as np
from tensorflow.python.client import device_lib
import tensorflow as tf

def scope_decorator(function):
    """
    Decorator that handles graph construction and variable scoping
    """

    name = function.__name__
    attribute = '_cache_' + name

    @property
    @functools.wraps(function)
    def decorator(self):
        if not hasattr(self,attribute):
            with tf.variable_scope(name):
                setattr(self,attribute,function(self))
        return getattr(self,attribute)

    return decorator

def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']

def weight_variable(shape, stddev):
    """
    Creates a variable tensor with a shape defined by the input
    Inputs:
        shape: list containing dimensionality of the desired output
        stddev: standard deviation to initialize with
    """
    initial = tf.truncated_normal(shape, stddev=stddev)
    return tf.Variable(initial)

def bias_variable(shape, value=0.1):
    """
    Creates a variable tensor with dimensionality defined by the input and
    initializes it to a constant
    Inputs:
        shape: list containing dimensionality of the desired output
        value: float specifying the initial value of the variable
    """
    initial = tf.constant(value, shape=shape)
    return tf.Variable(initial)

class SoSModel:
    def __init__(
                    self,
                    feature_size=41,
                    label_size=21,
                    rnn_size=100,
                ):
        """
        Initializes the model
        """

        self.feature_size = feature_size
        self.label_size = label_size
        self.rnn_size = rnn_size

        # Check for available gpus
        gpus = get_available_gpus()

        self.graph = tf.Graph()

        # Use gpu:0 if one is available
        if len(gpus) > 0:
            with self.graph.as_default():
                with tf.device(gpus[0]):
                    self._build_model()
        else:
            with self.graph.as_default():
                self._build_model()

        # Session config
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True

        # Create a session to run this graph
        self.sess = tf.Session(
                               config=config,
                               graph=self.graph
                              )

    def __del__(self):
        """
        Close the session when the model is deleted
        """

        self.sess.close()

    def initialize(self):
        """
        Initialize variables in the graph
        """

        with self.graph.as_default():
            self.sess.run(tf.global_variables_initializer())

    def _build_model(self):
        """
        Build the model graph
        """
        # Placeholder for learning rate
        self.lr = tf.placeholder(tf.float32)

        # Placeholder tensor for the input sessions
        self.X = tf.placeholder(
                                tf.float32,
                                [None, None, self.feature_size]
                               )

        # Placeholder tensor for the input representations
        self.L = tf.placeholder(tf.float32, [None,self.label_size])

        # Placeholder tensor for the labels/targets
        self.Y = tf.placeholder(tf.float32, [None])

        # Model methods
        self.network
        self.cost
        self.optimizer
        self.get_output

        # Saver
        self.saver = tf.train.Saver()

    @scope_decorator
    def network(self):
        """
        Construct the network used for classifying sessions
        """

        # Get the shape of the input
        shape = tf.shape(self.X)

        # Embed the session sequence with an LSTM with attention
        with tf.variable_scope('session_rnn', reuse=None):
            session_cell = tf.contrib.rnn.BasicLSTMCell(
                                                self.rnn_size,
                                                activation=tf.tanh
                                                         )

            session_vectors, session_states = tf.nn.dynamic_rnn(
                                                                session_cell,
                                                                self.X,
                                                                dtype=tf.float32
                                                               )
        # session_vectors should be (batch_size, time_step, rnn_size)
        # Reshape to (batch_size*time_step, rnn_size) to calc outputs
        vectors = tf.reshape(session_vectors, (shape[0]*shape[1],self.rnn_size))

        # Pass the vectors through a feedforward layer
        std_dev = np.sqrt(2)/np.sqrt(self.rnn_size+self.label_size)
        weights_1 = weight_variable([self.rnn_size,self.label_size],std_dev)
        biases_1 = bias_variable([self.label_size], value=0.0)
        layer_1 = tf.matmul(vectors,weights_1) + biases_1
        layer_1 = layer_1
        layer_1 = tf.reshape(layer_1, (shape[0],shape[1],self.label_size))

        # Compute the weighted average over the class vectors
        class_vectors = tf.expand_dims(self.L, axis=1)
        weighted_average = tf.reduce_sum(layer_1*class_vectors, axis=2)

        return weighted_average[:,-1], tf.sigmoid(weighted_average)

    @scope_decorator
    def cost(self):
        """
        Binary cross entropy cost
        """
        output, _ = self.network
        cost = tf.nn.sigmoid_cross_entropy_with_logits(
                                                        labels=self.Y,
                                                        logits=output
                                                      )
        return tf.reduce_mean(cost)

    @scope_decorator
    def optimizer(self):
        """
        Constructs the optimizer op used to train the network.
        Use gradient clipping.
        """
        opt = tf.train.AdamOptimizer()
        #gradients, variables = zip(*opt.compute_gradients(self.cost))
        #gradients, _ = tf.clip_by_global_norm(gradients, 0.1)
        #return opt.apply_gradients(zip(gradients, variables))
        return opt.minimize(self.cost)

    def save(self, path):
        """
        Saves the model to the specified path.
        """
        self.saver.save(self.sess, path)

    def load(self, path):
        """
        Load the model from the specified path.
        """
        self.saver.restore(self.sess, path)

    def train_on_batch(self, X, L, Y, learning_rate=0.01):
        """
        Train model for one step on specified batch
        """
        c, _ = self.sess.run([self.cost, self.optimizer],
                             {
                                self.X: X,
                                self.L: L,
                                self.Y: Y,
                                self.lr: learning_rate
                             }
                            )
        return c

    def get_cost(self, X, L, Y):
        """
        Get the cost on a specified batch
        """
        c = self.sess.run(
                            self.cost,
                                {
                                 self.X: X,
                                 self.L: L,
                                 self.Y: Y
                                }
                         )
        return c

    def get_output(self, X, L):
        """
        Get the predictions given input data
        """
        _, out = self.sess.run(
                                self.network,
                                    {
                                        self.X: X,
                                        self.L: L
                                    }
                              )

        return out



from setuptools import setup

setup(
    name='poseidonml',
    version='0.1.4',
    packages=['poseidonml'],
    package_dir={'poseidonml':'utils'},
    package_data={'poseidonml': ['models/*']},
    install_requires=['numpy==1.14.5', 'pika==0.12.0', 'redis==2.10.6', 'scikit-learn==0.18.2', 'scipy==1.1.0', 'tensorflow==1.9.0'],
    license='Apache License 2.0',
    author='cglewis',
    author_email='clewis@iqt.org',
    maintainer='Charlie Lewis',
    maintainer_email='clewis@iqt.org',
    description=('A utility package for extracting and analyzing features in network traffic.'),
    keywords='machine learning network analysis utilities',
    url='https://github.com/CyberReboot/PoseidonML',
)



import functools
import numpy as np
from tensorflow.python.client import device_lib
import tensorflow as tf

def scope_decorator(function):
    """
    Decorator that handles graph construction and variable scoping
    """

    name = function.__name__
    attribute = '_cache_' + name

    @property
    @functools.wraps(function)
    def decorator(self):
        if not hasattr(self,attribute):
            with tf.variable_scope(name):
                setattr(self,attribute,function(self))
        return getattr(self,attribute)

    return decorator

def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [x.name for x in local_device_protos if x.device_type == 'GPU']

def weight_variable(shape, stddev):
    """
    Creates a variable tensor with a shape defined by the input
    Inputs:
        shape: list containing dimensionality of the desired output
        stddev: standard deviation to initialize with
    """
    initial = tf.truncated_normal(shape, stddev=stddev)
    return tf.Variable(initial)

def bias_variable(shape, value=0.1):
    """
    Creates a variable tensor with dimensionality defined by the input and
    initializes it to a constant
    Inputs:
        shape: list containing dimensionality of the desired output
        value: float specifying the initial value of the variable
    """
    initial = tf.constant(value, shape=shape)
    return tf.Variable(initial)

class AbnormalDetector:
    def __init__(
                    self,
                    packet_embedding_size=100,
                    session_embedding_size=100,
                    hidden_size=100,
                    num_chars=16,
                    num_labels=32,
                    attn_size=10
                ):
        """
        Initializes the model
        """

        self.packet_embedding_size = packet_embedding_size
        self.session_embedding_size = session_embedding_size
        self.hidden_size = hidden_size
        self.num_chars = num_chars
        self.num_labels = num_labels
        self.attn_size = attn_size

        # Check for available gpus
        gpus = get_available_gpus()

        self.graph = tf.Graph()

        # Use gpu:0 if one is available
        if len(gpus) > 80:
            with self.graph.as_default():
                with tf.device(gpus[0]):
                    print("Using", gpu[0])
                    self._build_model()
        else:
            with self.graph.as_default():
                self._build_model()

        # Session config
        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True

        # Create a session to run this graph
        self.sess = tf.Session(
                               config=config,
                               graph=self.graph
                              )

    def __del__(self):
        """
        Close the session when the model is deleted
        """

        self.sess.close()

    def initialize(self):
        """
        Initialize variables in the graph
        """

        with self.graph.as_default():
            self.sess.run(tf.global_variables_initializer())

    def _build_model(self):
        """
        Build the model graph
        """
        # Placeholder for learning rate
        self.lr = tf.placeholder(tf.float32)

        # Placeholder tensor for the input sessions
        self.X = tf.placeholder(
                                tf.float32,
                                [None, None, 116, self.num_chars]
                               )

        # Placeholder tensor for the input representations
        self.R = tf.placeholder(tf.float32, [None,self.num_labels])

        # Placeholder tensor for the labels/targets
        self.Y = tf.placeholder(tf.int16, [None, 1])

        # Model methods
        self.network
        self.cost
        self.optimizer
        self.get_output

        # Saver
        self.saver = tf.train.Saver()

    @scope_decorator
    def network(self):
        """
        Construct the network used for classifying sessions
        """

        # Get the shape of the input
        shape = tf.shape(self.X)

        # Reshape the packet number into batch to embed each packet separately
        X_reshaped = tf.reshape(
                                self.X,
                                [shape[0]*shape[1],shape[2],self.num_chars]
                               )

        # Embed the packets with a BLSTM  with attention
        with tf.variable_scope('packet_rnn', reuse=None):
            # Attend over the input
            packet_cell_f = tf.contrib.rnn.BasicLSTMCell(
                                                self.packet_embedding_size//2,
                                                activation=tf.tanh
                                                        )
            packet_cell_b = tf.contrib.rnn.BasicLSTMCell(
                                                self.packet_embedding_size//2,
                                                activation=tf.tanh
                                                        )
            packet_attn_cell_f = tf.contrib.rnn.AttentionCellWrapper(
                                                packet_cell_f,
                                                attn_length=self.attn_size
                                                                    )
            packet_attn_cell_b = tf.contrib.rnn.AttentionCellWrapper(
                                                packet_cell_b,
                                                attn_length=self.attn_size
                                                                    )
            packet_vectors, packet_states = tf.nn.bidirectional_dynamic_rnn(
                                                            packet_attn_cell_f,
                                                            packet_attn_cell_b,
                                                            X_reshaped,
                                                            dtype=tf.float32
                                                                           )

        packet_vectors_f = packet_vectors[0][:,-1,:]
        packet_vectors_b = packet_vectors[1][:,-1,:]
        packet_vectors = tf.concat([packet_vectors_f, packet_vectors_b],1)

        # Reshape this to recover the session dimension
        sessions = tf.reshape(
                              packet_vectors,
                              [shape[0],shape[1],self.packet_embedding_size]
                             )

        # Embed the sessions with a BLSTM with attention
        with tf.variable_scope('session_rnn', reuse=None):
            # Attend over the packets
            session_cell_f = tf.contrib.rnn.BasicLSTMCell(
                                                self.session_embedding_size//2,
                                                activation=tf.tanh
                                                         )
            session_cell_b = tf.contrib.rnn.BasicLSTMCell(
                                                self.session_embedding_size//2,
                                                activation=tf.tanh
                                                         )
            session_attn_cell_f = tf.contrib.rnn.AttentionCellWrapper(
                                                    session_cell_f,
                                                    attn_length=self.attn_size
                                                                     )
            session_attn_cell_b = tf.contrib.rnn.AttentionCellWrapper(
                                                    session_cell_b,
                                                    attn_length=self.attn_size
                                                                     )
            session_vectors, session_states = tf.nn.bidirectional_dynamic_rnn(
                                                            session_attn_cell_f,
                                                            session_attn_cell_b,
                                                            sessions,
                                                            dtype=tf.float32
                                                                             )

        session_vectors_f = session_vectors[0][:,-1,:]
        session_vectors_b = session_vectors[1][:,-1,:]
        session_vectors = tf.concat([session_vectors_f, session_vectors_b],1)

        # Pass the RNN output through a feedforward layer
        std_dev = np.sqrt(2)/np.sqrt(
                                self.session_embedding_size+self.hidden_size
                                    )
        weights_1 = weight_variable(
                                [self.session_embedding_size,self.hidden_size],
                                std_dev
                                   )
        biases_1 = bias_variable([self.hidden_size], value=0.0)
        layer_1 = tf.matmul(session_vectors,weights_1) + biases_1
        layer_1 = tf.nn.tanh(layer_1)

        # Compute the weighted average over the class vectors
        std_dev = np.sqrt(2)/np.sqrt(self.hidden_size + 1)
        class_vectors = weight_variable(
                                        [self.hidden_size,self.num_labels],
                                        std_dev
                                       )
        class_biases = bias_variable([self.num_labels],value=0.0)

        R = tf.expand_dims(self.R, axis=1)
        classes = tf.expand_dims(class_vectors, axis=0)
        biases = tf.expand_dims(class_biases,axis=0)

        RC = R*classes
        RB = self.R*biases

        weighted_classes = tf.reduce_sum(RC, axis=2)
        weighted_bias = tf.reduce_sum(RB,axis=1)

        output = weighted_classes * layer_1
        output = tf.reduce_sum(output, axis=1)
        output = output + weighted_bias
        output = tf.expand_dims(output, axis=1)
        probs = tf.nn.sigmoid(output)

        return output, probs

    @scope_decorator
    def cost(self):
        """
        Binary cross entropy cost
        """
        output, _ = self.network
        cost = tf.losses.sigmoid_cross_entropy(self.Y, output)
        return cost

    @scope_decorator
    def optimizer(self):
        """
        Constructs the optimizer op used to train the network.
        Use gradient clipping.
        """
        opt = tf.train.AdamOptimizer()
        gradients, variables = zip(*opt.compute_gradients(self.cost))
        gradients, _ = tf.clip_by_global_norm(gradients, 0.01)
        return opt.apply_gradients(zip(gradients, variables))

    def save(self, path):
        """
        Saves the model to the specified path.
        """
        self.saver.save(self.sess, path)

    def load(self, path):
        """
        Load the model from the specified path.
        """
        self.saver.restore(self.sess, path)

    def train_on_batch(self, X, R, Y, learning_rate=0.01):
        """
        Train model for one step on specified batch
        """
        c, _ = self.sess.run([self.cost, self.optimizer],
                             {
                                self.X: X,
                                self.R: R,
                                self.Y: Y,
                                self.lr: learning_rate
                             }
                            )
        return c

    def get_cost(self, X, R, Y):
        """
        Get the cost on a specified batch
        """
        c = self.sess.run(
                            self.cost,
                                {
                                 self.X: X,
                                 self.R: R,
                                 self.Y: Y
                                }
                         )
        return c

    def get_output(self, X, R):
        """
        Get the predictions given input data
        """
        _, out = self.sess.run(
                                self.network,
                                    {
                                        self.X: X,
                                        self.R: R
                                    }
                              )

        return out



import numpy as np
from collections import defaultdict
from .pcap_utils import extract_macs, \
                       is_private, \
                       is_external, \
                       is_protocol, \
                       get_source, \
                       get_ip_port
import json

def extract_features(session_dict, capture_source=None, max_port=None):
    '''
    Extracts netflow level features from packet capture.

    Args:
        pcap_path: path to the packet capture to process into features
        max_port:  Maximum port to get features on (default to reading config)

    Returns:
        feature_vector: Vector containing the featurized representation
                        of the input pcap.
    '''

    # Get featurization info from config
    try:
        with open('opts/config.json', 'r') as config_file:
            config = json.load(config_file)
            address_type = config['source identifier']
            if max_port is None:
                max_port = config['max port']
    except:
        address_type = 'MAC'

    # If the capture source isn't specified, default to the most used address
    if capture_source is None:
        capture_source = get_source(session_dict, address_type=address_type)

    # Initialize some counter variables
    num_sport_init = [0]*max_port
    num_dport_init = [0]*max_port
    num_sport_rec = [0]*max_port
    num_dport_rec = [0]*max_port

    num_sessions_init = 0
    num_external_init = 0
    num_tcp_sess_init = 0
    num_udp_sess_init = 0
    num_icmp_sess_init = 0

    num_sessions_rec = 0
    num_external_rec = 0
    num_tcp_sess_rec = 0
    num_udp_sess_rec = 0
    num_icmp_sess_rec = 0

    # Iterate over all sessions and aggregate the info
    other_ips = defaultdict(int)
    for key, session in session_dict.items():
        address_1, port_1 = get_ip_port(key[0])
        address_2, port_2 = get_ip_port(key[1])
        
        # Get the first packet and grab the macs from it
        first_packet = session[0][1]
        source_mac, destination_mac = extract_macs(first_packet)

        # If the source is the cpature source
        if (source_mac == capture_source
            or address_1 == capture_source):

            if is_private(address_2):
                other_ips[address_2] += 1

            num_sessions_init += 1
            num_external_init += is_external(address_1, address_2)
            num_tcp_sess_init += is_protocol(session, '06')
            num_udp_sess_init += is_protocol(session, '11')
            num_icmp_sess_init += is_protocol(session, '01')

            if int(port_1) < max_port:
                num_sport_init[int(port_1)] += 1

            if int(port_2) < max_port:
                num_dport_init[int(port_2)] += 1

        # If the destination is the capture source
        if (destination_mac == capture_source
            or address_2 == capture_source):
            if is_private(address_1):
                other_ips[address_1] += 1

            num_sessions_rec += 1
            num_external_rec += is_external(address_2, address_1)
            num_tcp_sess_rec += is_protocol(session, '06')
            num_udp_sess_rec += is_protocol(session, '11')
            num_icmp_sess_rec += is_protocol(session, '01')

            if int(port_1) < max_port:
                num_sport_rec[int(port_1)] += 1
            if int(port_2) < max_port:
                num_dport_rec[int(port_2)] += 1

    num_port_sess = np.concatenate(
                                   (
                                   num_sport_init,
                                   num_dport_init,
                                   num_sport_rec,
                                   num_dport_rec
                                   ),
                                    axis=0
                                  )

    if num_sessions_init == 0:
        num_sessions_init += 1
    if num_sessions_rec == 0:
        num_sessions_rec += 1

    num_port_sess=np.asarray(num_port_sess)/(num_sessions_init+num_sessions_rec)

    extra_features = [0]*8
    extra_features[0] = num_external_init/num_sessions_init
    extra_features[1] = num_tcp_sess_init/num_sessions_init
    extra_features[2] = num_udp_sess_init/num_sessions_init
    extra_features[3] = num_icmp_sess_init/num_sessions_init

    extra_features[4] = num_external_rec/num_sessions_rec
    extra_features[5] = num_tcp_sess_rec/num_sessions_rec
    extra_features[6] = num_udp_sess_rec/num_sessions_rec
    extra_features[7] = num_icmp_sess_rec/num_sessions_rec

    feature_vector = np.concatenate((num_port_sess, extra_features), axis=0)
    return feature_vector, capture_source, list(other_ips.keys())



'''
Utilities for preparing sessions for input into models
'''
from collections import OrderedDict, defaultdict, Counter
import numpy as np
import os

def is_private(address):
    '''
    Checks if an address is private and if so returns True.  Otherwise returns
    False.
    Args:
        address: Address to check. Can be list or string
    Returns:
        True or False
    '''
    if '.' in address:  # ipv4
        pairs = address.split('.')
    elif ':' in address:  # ipv6
        pairs = address.split(':')
    else:  # unknown
        pairs = []

    private = False
    if pairs:
        if pairs[0] == '10': private = True
        elif pairs[0] == '192' and pairs[1] == '168': private = True
        elif pairs[0] == '172' and 16 <= int(pairs[1]) <= 31: private = True
        elif pairs[0] == 'fe80': private = True
        elif pairs[0].startswith('fd'): private = True

    return private

def extract_macs(packet):
    '''
    Takes in hex representation of a packet header and extracts the
    source and destination mac addresses

    returns:
        source_mac: Destination MAC address
        destination_mac: Destination MAC address
    '''

    source_mac = packet[12:24]
    dest_mac = packet[0:12]

    source_mac = ':'.join(source_mac[i:i+2]
                          for i in range(0,len(source_mac), 2)
                         )
    destination_mac = ':'.join(dest_mac[i:i+2]
                               for i in range(0,len(dest_mac),2)
                              )

    return source_mac, destination_mac

def get_indiv_source(sessions, address_type='MAC'):
    '''
    Gets the source MAC address from an individual session dictionary.
    Also computes the number of sessions to and from this source.
    The source is defined to be the IP address with the most sessions
    associated with it.

    Inputs:
        sessions: A dictionary of hex sessions from the sessionizer
        address_type: Type of address to report as the source
    Returns:
        capture_source: Address of the capture source
        ip_mac_pairs: Counts of appearances of ip:mac pairs
    '''

    # Number of sessions involving the address
    ip_mac_pairs = defaultdict(int)

    # Count the incoming/outgoing sessions for all addresses
    for key in sessions:
        source_address, _ = get_ip_port(key[0])
        destination_address, _ = get_ip_port(key[1])

        # Get the first packet and grab the macs from it
        first_packet = sessions[key][0][1]
        source_mac, destination_mac = extract_macs(first_packet)

        # Compute the IP/MAC address pairs
        if os.environ.get('POSEIDON_PUBLIC_SESSIONS'):
            pair_1 = source_address + '-' + source_mac
            pair_2 = destination_address + '-' + destination_mac
            ip_mac_pairs[pair_1] += 1
            ip_mac_pairs[pair_2] += 1
        else:
            # Only look at sessions with an internal IP address
            # This shouldn't actually be necessary at this stage
            if is_private(source_address) or is_private(destination_address):
                pair_1 = source_address + '-' + source_mac
                pair_2 = destination_address + '-' + destination_mac
                if is_private(source_address):
                    ip_mac_pairs[pair_1] += 1
                if is_private(destination_address):
                    ip_mac_pairs[pair_2] += 1

    # The address with the most sessions is the capture source
    if len(sessions) == 0:
        return None, ip_mac_pairs

    sorted_sources = sorted(
                            ip_mac_pairs.keys(),
                            key=(lambda k: ip_mac_pairs[k]),
                            reverse=True
                           )
    if address_type == 'MAC':
        capture_source = '00:00:00:00:00:00'
    else:
        capture_source = '0.0.0.0'

    if len(sorted_sources) > 0:
        if address_type == 'MAC':
            capture_source = sorted_sources[0].split('-')[1]
        else:
            capture_source = sorted_sources[0].split('-')[0]

    return capture_source, ip_mac_pairs

def get_source(sessions, address_type='MAC'):
    '''
    Gets the source MAC for all the session dicts given.  This is the majority
    vote across all session dicts if sessions is a list.

    Args:
        sessions: either a single session dict or a list of session dicts
        address_type: Type of address to return as source identifer
    Returns
        capture_source: Majority MAC address across all sessions in input
    '''

    if type(sessions) is list:
        all_pairs = Counter({})
        # Aggregate counts from all binned sessions
        if address_type == 'MAC':
            capture_source = '00:00:00:00:00:00'
        else:
            capture_source = '0.0.0.0'

        for session_dict in sessions:
            # Get the ip mac address pairs for each session dict
            _, ip_mac_pairs = get_indiv_source(session_dict)
            # Combine with previous stats
            all_pairs += Counter(ip_mac_pairs)

        # Find the most common address
        sorted_sources = sorted(
                                all_pairs.keys(),
                                key=(lambda k: all_pairs[k]),
                                reverse=True
                               )

        if len(sorted_sources) > 0:
            if address_type == 'MAC':
                capture_source = sorted_sources[0].split('-')[1]
            else:
                capture_source = sorted_sources[0].split('-')[0]

    else:
        if address_type == 'MAC':
            capture_source, _ = get_indiv_source(sessions)
        else:
            capture_source, _ = get_indiv_source(sessions, address_type='IP')

    return capture_source

def packet_size(packet):
    '''
    Extracts the size of a packet in bytes from the hex header.

    Args:
        packet: Hex header of the packet

    Returns:
        size: Size in bytes of the IP packet, including data
    '''

    size = packet[1][32:36]
    try:
        size = int(size, 16)
    except:
        size = 0

    return size

def extract_session_size(session):
    '''
    Extracts the total size of a session in bytes.

    Args:
        session: session list containing all the packets of the session

    Returns:
        session_size: Size of the session in bytes
    '''

    session_size = sum([packet_size(p) for p in session])
    return session_size

def extract_protocol(session):
    '''
    Extracts the protocol used in the session from the first packet

    Args:
        session: session tuple containing all the packets of the session

    Returns:
        protocol: Protocol number used in the session
    '''

    protocol = session[0][1][46:48]
    return protocol

def is_external(address_1, address_2):
    '''
    Checks if a session is between two sources within the same network.
    For now this is defined as two IPs with the first octet matching.

    Args:
        address_1: Address of source participant
        address_2: Address of destination participant

    Returns:
        is_external: True or False if this is an internal session
    '''

    if is_private(address_1) and is_private(address_2):
        return False

    return True

def is_protocol(session, protocol):
    '''
    Checks if a session is of the type specified

    Args:
        session: List of packets in the session
        protocol: Protocol to check

    Returns:
        is_protocol: True or False indicating if this is a TCP session
    '''

    p = extract_protocol(session)
    if protocol == p:
        return True
    return False

def strip_macs(packet):
    '''
    Strip the mac addresses out of a packet
    '''
    return packet[24:]

def strip_ips(stripped_packet):
    '''
    Strip the IP addresses out of a packet that has had its mac addresses
    stripped out
    '''
    return stripped_packet[0:28] + stripped_packet[44:]

def clean_packet(packet):
    '''
    Remove both mac and ip addresses from a packet
    '''
    no_macs = strip_macs(packet)
    no_ips = strip_ips(no_macs)
    return no_ips

def clean_session_dict(sessions, source_address=None):
    '''
    return sessions of packets with no mac or ip addresses from the source
    '''
    if source_address is None:
        source_address = get_source(sessions, address_type='IP')

    def clean_dict(sessions, source_address):
        cleaned_sessions = OrderedDict()
        for key, packets in sessions.items():
            address_1, port_1 = get_ip_port(key[0])
            address_2, port_2 = get_ip_port(key[1])

            first_packet = sessions[key][0][1]
            source_mac, destination_mac = extract_macs(first_packet)

            if (address_1 == source_address
                or source_mac == source_address
                or address_2 == source_address
                or destination_mac == source_address):
                if os.environ.get('POSEIDON_PUBLIC_SESSIONS'):
                    cleaned_sessions[key] = [
                                             (ts, clean_packet(p))
                                             for ts, p in packets[0:8]
                                            ]
                else:
                    if is_private(address_1) or is_private(address_2):
                        cleaned_sessions[key] = [
                                                 (ts, clean_packet(p))
                                                 for ts, p in packets[0:8]
                                                ]
        return cleaned_sessions

    if type(sessions) == list:
        cleaned_sessions = []
        for sess in sessions:
            cleaned_sessions.append(clean_dict(sess,source_address))
    else:
        cleaned_sessions = clean_dict(sessions, source_address)

    return cleaned_sessions, source_address

def create_inputs(labels, session, seq_len, num_chars=16):
    '''
    Creates model inputs from a set of labels session
    '''
    L = np.zeros((1, len(labels)))
    X = np.zeros((1, len(session), seq_len, num_chars))

    # Create the packet input
    hex_str = '0123456789abcdef'
    for i, key in enumerate(session):
        raw_hex = session[i][1]
        for j, c in enumerate(raw_hex):
            if j < seq_len:
                char_id = hex_str.index(c)
                X[0,i,j,char_id] = 1

    # Create the label input
    classification = sorted(labels, key=lambda x: x[0])
    class_array = [p for c,p in classification]
    L[0] = np.asarray(class_array)

    return X, L

def get_length(packet):
    """
    Gets the total length of the packet
    """
    hex_str = '0123456789abcdef'
    hex_length = packet[32:36]
    length = 0
    for i,c in enumerate(hex_length[::-1]):
        length += pow(16,i)*hex_str.index(c)
    return length


def featurize_session(key, packets, source=None):
    # Global session properties
    address_1, port_1 = get_ip_port(key[0])
    address_2, port_2 = get_ip_port(key[1])
    if address_1 == source or address_2 == source or source == None:
        initiated_by_source = None
        if address_1 == source:
            initiated_by_source = True
        if address_2 == source:
            initiated_by_source = False

        mac_1, mac_2 = extract_macs(packets[0][1])
        protocol = extract_protocol(packets)
        external = is_external(address_1, address_2)

        # Packet specific properties
        size_to_1 = 0
        size_to_2 = 0

        first_time = packets[0][0].timestamp()
        last_time = packets[-1][0].timestamp()

        num_sent_by_1 = 0
        num_sent_by_2 = 0
        for packet in packets:
            time = packet[0].timestamp()
            source_mac, destination_mac = extract_macs(packet[1])

            if source_mac == mac_1:
                size_to_2 = get_length(packet[1])
                num_sent_by_1 += 1
            if source_mac == mac_2:
                size_to_1 = get_length(packet[1])
                num_sent_by_2 += 1
        if (num_sent_by_1 + num_sent_by_2) > 1:
            freq_1 = num_sent_by_1/(last_time - first_time)
            freq_2 = num_sent_by_2/(last_time - first_time)
        else:
            freq_1 = 1
            freq_2 = 1

        # Netflow-like session info
        session_info = {
                        'start time': packets[0][0],
                        'initiated by source': initiated_by_source,
                        'external session': external,
                        'source': key[0],
                        'destination': key[1],
                        'protocol': protocol,
                        'data to source': size_to_1,
                        'data to destination': size_to_2,
                        'packets to source': num_sent_by_2,
                        'packets to destination': num_sent_by_1,
                        'source frequency': freq_1,
                        'destination frequency': freq_2,
                       }
        return session_info
    else:
        return None


def get_ip_port(socket_str):
    """
    Returns ip and port
    :param socket_str: ipv4/6:port
    :return:
    address, port
    """
    splitter_index = socket_str.rindex(':')
    address = socket_str[0:splitter_index]
    port = socket_str[splitter_index + 1:]

    return address, port



import os
import sys
import json
import logging
import pickle
import numpy as np
import tensorflow as tf
from pkg_resources import working_set
from pkg_resources import Requirement
from .SoSmodel import SoSModel
from .session_sequence import create_dataset
from .session_iterator import BatchIterator
import time

logging.basicConfig(level=logging.INFO)
tf.logging.set_verbosity(tf.logging.ERROR)

# Load info from config
with open('opts/config.json') as config_file:
    config = json.load(config_file)
    rnn_size = config['rnn size']
    labels = config['labels']

def eval_pcap(pcap, label=None):
    logger = logging.getLogger(__name__)
    data = create_dataset(pcap, label=label)
    # Create an iterator
    iterator = BatchIterator(
                             data,
                             labels,
                             perturb_types=['random data']
                            )
    logger.debug("Created iterator")
    rnnmodel = SoSModel(rnn_size=100)
    logger.debug("Created model")
    rnnmodel.load(os.path.join(working_set.find(Requirement.parse('poseidonml')).location, 'poseidonml/models/SoSmodel'))
    logger.debug("Loaded model")

    X_list = iterator.X
    L_list = iterator.L
    sessions = iterator.sessions

    num_total = 0
    num_abnormal = 0
    max_score = 0
    scores = {}
    for i, X in enumerate(X_list):
        L = L_list[i]
        out = rnnmodel.get_output(
                                    np.expand_dims(X, axis=0),
                                    np.expand_dims(L, axis=0),
                                 )
        for j,o in enumerate(out):
            for k,s in enumerate(o):
                num_total += 1
                session = sessions[i][k]['session info']
                p = session['protocol']
                if p == '06': p = 'TCP'
                if p == '17': p = 'UDP'
                if p == '01': p == 'ICMP'
                flowlike = p + ' '
                if session['initiated by source']:
                    flowlike += session['source']+' to '+session['destination']
                else:
                    flowlike += session['destination']+' to '+session['source']
                scores[num_total] = str(s)
                if s > max_score:
                    max_score = s

    '''
    print("Processed",num_total,"sessions of which",num_abnormal,"were abnormal")
    with open('sos_output.json', 'w') as output_file:
        json.dump(scores, output_file)
    '''
    return max_score

if __name__ == '__main__':
    # Path to training data
    pcap = sys.argv[1]
    if len(sys.argv) == 3:
        label = sys.argv[2]
    else:
        label = None
    mean_score = eval_pcap(pcap,label=label)
    print(mean_score)



"""
Contains iterator class for generating training batches from a canned dataset
"""

import pickle
import numpy as np
from sklearn.model_selection import train_test_split

class BatchIterator:
    def __init__(
                 self,
                 data_input,
                 labels,
                 batch_size=64,
                 seq_len=10,
                 ports=None,
                 perturb_types="all",
                 seed=0
                ):
        """
        Initialize the iterator with specified hyperparameters and load the
        data from the specified path
        """
        self.rng = np.random.RandomState(seed)
        self.data_input = data_input

        self.labels = sorted(labels)
        self.classification_length = len(labels)
        self.seq_len = seq_len
        self.data = None
        self.X = []
        self.L = []
        self.means = None
        self.stds = None
        self.sessions = []

        if ports is None:
            self.ports = [22,53,67,68,80,88,123,135,137,138,192,389,443,445,514]
        else:
            self.ports = ports
        self.feature_length = 11 + 2*len(self.ports)
        self._load_data()

        self.X_train, self.X_test, self.L_train, self.L_test = train_test_split(
                                                                self.X,
                                                                self.L,
                                                                test_size=0.2,
                                                                random_state=0
                                                                               )

        self.X_vala, self.X_test, self.L_vala, self.L_test = train_test_split(
                                                                self.X_test,
                                                                self.L_test,
                                                                test_size=0.5,
                                                                random_state=0
                                                                             )

        self.X_train = np.array(self.X_train)
        self.X_test = np.array(self.X_test)
        self.X_vala = np.array(self.X_vala)
        self.train_length = self.X_train.shape[0]
        self.validation_length = self.X_vala.shape[0]
        self.test_length = self.X_test.shape[0]
        #self._normalize()

        self.perturb_types = perturb_types

    def _load_data(self):
        """
        Handles loading the data into the correct format
        """
        if type(self.data_input) is dict:
            self.data = self.data_input
        else:
            with open(self.data_input,'rb') as handle:
                data = pickle.load(handle)
            self.data = data

        for pcap, sessions in self.data.items():
            x = np.zeros((self.seq_len, self.feature_length))
            l = np.zeros((self.classification_length, self.seq_len))
            i = 0
            session_list = []
            for session in sessions:
                session_list.append(session)
                x_sess, l_sess = self._vectorize(session)
                x[i] = x_sess
                l[:,i] = l_sess
                i += 1
                if i >= self.seq_len:
                    self.X.append(x)
                    self.L.append(np.mean(l, axis=1))
                    self.sessions.append(session_list)
                    x = np.zeros((self.seq_len, self.feature_length))
                    l = np.zeros((self.classification_length, self.seq_len))
                    i = 0
        if self.X and self.L:
            self.X = np.stack(self.X)
            self.L = np.stack(self.L)
            self.data_length = self.X.shape[0]

    def _vectorize(self, session):
        '''
        Turns a session into a feature vector
        '''

        X = np.zeros(11)
        y = np.zeros(self.classification_length)

        session_info = session['session info']

        source_port = int(session_info['source'].split(':')[1])
        destination_port = int(session_info['destination'].split(':')[1])


        # Feature for if the source initated the session
        if session_info['initiated by source']:
            X[0] = 1

        # Feature for external/internal session
        if session_info['external session']:
            X[1] = 1

        # One hot feature for the protocol
        if session_info['protocol'] == '01': # For ICMP
            X[2] = 1
        elif session_info['protocol'] == '06': # For TCP
            X[3] = 1
        elif session_info['protocol'] == '11': # For UDP
            X[4] = 1

        # Feature for amount of data sent to source/destination
        X[5] = session_info['data to source']
        X[6] = session_info['data to destination']

        # Feature for packets sent to souce/destination
        X[7] = session_info['packets to source']
        X[8] = session_info['packets to destination']

        # Feature for frequencies of source and destination
        X[9] = session_info['source frequency']
        X[10] = session_info['destination frequency']

        # Vectorize port based features
        source_ports = np.zeros(len(self.ports))
        destination_ports = np.zeros(len(self.ports))
        if source_port in self.ports and X[0] == 1:
            source_ports[self.ports.index(source_port)] = 1
        if source_port in self.ports and X[0] == 0:
            destination_ports[self.ports.index(source_port)] = 1
        if destination_port in self.ports and X[0] == 0:
            source_ports[self.ports.index(destination_port)] = 1
        if destination_port in self.ports and X[0] == 1:
            destination_ports[self.ports.index(destination_port)] = 1

        port_vec = np.concatenate([source_ports,destination_ports])
        X = np.concatenate([X,port_vec])
        # Create the labels:
        for c in session['model outputs']['classification']:
            y[self.labels.index(c[0])] = c[1]

        return X, y

    def _normalize(self):
        means = np.mean(self.X_train, axis=(0,1))
        stds = np.std(self.X_train, axis=(0,1))

        means[0:5] = 0
        means[11:] = 0
        stds[0:5] = 1
        stds[11:] = 1

        self.means = means
        self.stds = stds

    def _swap_ports(self, X):
        '''
        Swaps ports in a single session in the sequence
        '''
        s = np.random.choice(X.shape[0])
        port_len = len(self.ports)
        X_swapped = np.copy(X)
        X_swapped[s,11:port_len+11], X_swapped[s,11+port_len:] = \
        X[s,11+port_len:], X[s,11:port_len+11]
        return X_swapped

    def _switch_host(self, X):
        '''
        Switch the directionality of sessions in the sequence
        '''
        X_switched = np.copy(X)
        for i in range(X.shape[0]):
            if X[i,0] == 0:
                X_switched[i,0] = 1
            elif X[i,0] == 1:
                X_switched[i,0] = 0
        return X_switched

    def _random_data(self, X):
        '''
        Generates a sequence of random data
        '''
        X_randomized = np.copy(X)
        np.random.shuffle(X_randomized)
        for i in range(X.shape[0]):
            X_randomized[i,0] = np.random.choice([0,1])
            X_randomized[i,1] = np.random.choice([0,1])

            proto = np.random.choice([2,3,4])
            X_randomized[i, 2:5] = np.zeros(3)
            X_randomized[i, proto] = 1

            port_len = len(self.ports)
            X_randomized[i,11:port_len+11] = np.zeros(port_len)
            X_randomized[i,port_len+11:] = np.zeros(port_len)
            src = np.random.choice(range(port_len))
            dst = np.random.choice(range(port_len))
            X_randomized[i,11+src] = 1
            X_randomized[i,11+port_len+dst] = 1
        return X_randomized

    def gen_data(self, split='train', batch_size=64, perturb=False):
        """
        Generates perturbed or unperturbed batches
        """
        if split == 'validation':
            X = self.X_vala
            L = self.L_vala
            length = self.validation_length
        elif split == 'test':
            X = self.X_test
            L = self.L_test
            length = self.test_length
        elif split == 'train':
            X = self.X_train
            L = self.L_train
            length = self.train_length

        X_list = []
        L_list = []
        for i in range(batch_size):
            idx = np.random.choice(range(length))
            X_chosen = X[idx]
            #X_chosen -= self.means
            #X_chosen /= self.stds

            if perturb is True:
                perturbation = np.random.choice(self.perturb_types)
                if perturbation == 'port swap':
                    X_chosen = self._swap_ports(X_chosen)
                if perturbation == 'direction swap':
                    X_chosen = self._switch_host(X_chosen)
                if perturbation == 'random data':
                    X_chosen = self._random_data(X_chosen)
            X_list.append(X_chosen)
            L_list.append(L[idx])

        return np.stack(X_list), np.stack(L_list)

    def gen_batch(self, split='train', batch_size=64):
        """
        Generate mixed batches of perturbed and unperturbed data
        """
        X_n, L_n = self.gen_data(
                                    split=split,
                                    batch_size=batch_size
                                )
        X_a, L_a = self.gen_data(
                                    split=split,
                                    batch_size=batch_size,
                                    perturb=True
                                )
        normals = np.zeros(X_n.shape[0])
        abnormals = np.ones(X_a.shape[0])
        X = np.concatenate([X_n, X_a])
        L = np.concatenate([L_n, L_a])
        y = np.concatenate([normals, abnormals])

        return X, L, y
